"""
Basic statistical inference (:mod:`mqr.inference.nonparametric`)
================================================================

.. :currentmodule:`mqr.inference.nonparametric`

User guide
    :doc:`/user_guide/inference`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

.. rubric:: Non-parametric statistics

.. autosummary::
    :template: autosummary/submodule.rst
    :toctree: generated/

    correlation
    dist
    median
    quantile
    variance
"""

import mqr.inference.nonparametric.correlation
import mqr.inference.nonparametric.dist
import mqr.inference.nonparametric.median
import mqr.inference.nonparametric.quantile
import mqr.inference.nonparametric.variance
